import * as Mqtt from 'mqtt'
import * as Mongo from 'mongodb'
import * as Queue from './queue'
import * as MqttMessage from './mqtt_message'


var console = require('chalk-console')


var MQTT_BROKER_URL : String = 'mqtt://localhost'
var MONGO_DATABASE_URL : String = 'mongodb://localhost/test'
var COLLECTION : String = 'mqttMongo'
var TOPIC : String = 'mqttMongo'

import options = require('commander')
options
  .version('0.0.1')
  .option('-b, --broker [broker url]', 'Broker URL', MQTT_BROKER_URL)
  .option('-m, --database [database url]', 'MongoDB Database URL', MONGO_DATABASE_URL)
  .option('-c, --collection [mongodb collection]', 'Collection in MongoDB Database', COLLECTION)
  .option('-t, --topic [topic to subscribe]', 'Topic to subscribe', TOPIC)
  .parse(process.argv)



var dbHandle : Mongo.Db;
var dbReady : boolean = false;

Mongo.MongoClient.connect(options['database'], 
{
  'server': {
    'reconnectTries': 5,
     'reconnectInterval': 1000,
     'socketOptions': {
       'autoReconnect': false
     }
   }  
})
  .then(
    (tmpDbHandle: Mongo.Db) => { 
      dbHandle = tmpDbHandle
      dbReady = true; 
      console.info("Connected to database")
      dbHandle.on('reconnectFailed', (err : any) => { console.warn(`Error on database ${err}`) })
      dbHandle.on('reconnect', () => { 
        console.info("Reconnect on database") 
        dbReady = true
        queue.knock()
      })
      dbHandle.on('timeout', () => { console.warn("Timeout on database") })
      dbHandle.on('close', () => { 
        console.info("Close on database") 
        dbReady = false
      })
    },
    (err: String) => {
      console.error("Unable to connect to database: %s", err)
      process.exit(1)
    }
)


var queue = new Queue.Queue<MqttMessage.MqttMessage>()
queue.on('data', () => {
  if (dbReady) { 
    while (! queue.isEmpty()) {
      var msg : MqttMessage.MqttMessage = queue.deq()
      console.info(`Something in the queue: ${JSON.stringify(msg)}`)
      var coll = dbHandle.collection(options['collection'])
      coll.insertOne(msg.getMessage())
        .then(
          (r) => { 
            console.success(`Successfully inserted into database ${JSON.stringify(msg.getMessage())}`)
          },
          (err) => { 
            console.error(`Error when trying to insert into database ${err}`)
            if (! dbReady) {
              console.info("Error occured while database connection is lost, re-enqueue msg.")
              queue.reenq(msg)
            } else {
              console.error(`Message ${JSON.stringify(msg.getMessage())} is lost`)
            }
          }
        )
    }
  } else {
    // console.info("Database currently not available, not reading from stream")
  }
})


var mqttClient = Mqtt.connect(options['broker'])
mqttClient.on('offline', () => { console.warn("MQTT client is offline") })
mqttClient.on('reconnect', () => { console.warn("MQTT client is reconnecting") })
mqttClient.on('close', () => { console.warn("MQTT connection closed") })

mqttClient.on('connect', () => {
  console.info("Connected to MQTT broker")
  mqttClient.subscribe(options['topic'])
  mqttClient.subscribe('MqttMongo/Command')
  mqttClient.publish('MqttMongo/Status', 'hello, started up')
})

var msgCnt : number = 0
mqttClient.on('message', (topic : string, message : string) => {
  msgCnt++;
  console.info(`Message received ${msgCnt}, topic ${topic}, payload ${message}`)

  if (topic == "MqttMongo/Command" && message == "shutdown") {
    console.info("Shutting down MqttMongo")
    clearInterval(uptimeInterval)
    mqttClient.end()
    dbHandle.close()
  } else {
    try {
      var mqttMessage = new MqttMessage.MqttMessage(topic, message)
      queue.enq(mqttMessage)
    } catch (e) {
      console.error(`Error while parsing Mqtt message, topic '${topic}', message '${message}' , ${e.toString()}`)
    }
  }
})


var uptime : number = 0
var uptimeInterval = setInterval(() => {
  uptime++
  mqttClient.publish('MqttMongo/Status', `{'Uptime': ${uptime}}`)
}, 1000)



console.info("MqttMongo started")